/*
 * Copyright notice from original mutt:
 * Copyright (C) 1996,1997 Michael R. Elkins <me@mutt.org>
 * Copyright (C) 1999-2000 Thomas Roessler <roessler@does-not-exist.org>
 * Copyright (C) 2001  Thomas Roessler <roessler@does-not-exist.org>
 *                     Oliver Ehli <elmy@acm.org>
 * Copyright (C) 2003  Werner Koch <wk@gnupg.org>
 * Copyright (C) 2004 g10code GmbH
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/wait.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>

#ifdef HAVE_LOCALE_H
#  include <locale.h>
#endif
#ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
#endif
#ifdef HAVE_SYS_RESOURCE_H
#  include <sys/resource.h>
#endif

#include <lib-lib/lib-lib.h>

#include <lib-mime/mime.h>

#include <lib-ui/curses.h>

#include "mutt.h"
#include "alias.h"
#include "handler.h"
#include "copy.h"
#include "crypt.h"
#include "pgp.h"

/* print the current time to avoid spoofing of the signature output */
void crypt_current_time (STATE * s, const char *app_name)
{
  time_t t;
  char p[STRING], tmp[STRING];

  if (option (OPTCRYPTTIMESTAMP)) {
    t = time (NULL);
    setlocale (LC_TIME, "");
    strftime (p, sizeof (p), _(" (current time: %c)"), localtime (&t));
    setlocale (LC_TIME, "C");
  }
  else
    *p = '\0';

  snprintf (tmp, sizeof (tmp), _("[-- %s output follows%s --]\n"),
            NONULL (app_name), p);
  state_attach_puts (tmp, s);
}



void crypt_forget_passphrase (void)
{
    crypt_pgp_void_passphrase ();
    crypt_smime_void_passphrase ();
    mutt_message _("Passphrase(s) forgotten.");
}


#if defined(HAVE_SETRLIMIT)

static void disable_coredumps (void)
{
  struct rlimit rl = { 0, 0 };
  static short done = 0;

  if (!done) {
    setrlimit (RLIMIT_CORE, &rl);
    done = 1;
  }
}

#endif /* HAVE_SETRLIMIT */


int crypt_valid_passphrase (int flags)
{
  int ret = 0;

# if defined(HAVE_SETRLIMIT)
  disable_coredumps ();
# endif

  if (flags & APPLICATION_PGP)
    ret = crypt_pgp_valid_passphrase ();

  if (flags & APPLICATION_SMIME)
    ret = crypt_smime_valid_passphrase ();

  return ret;
}



int mutt_protect (HEADER * msg, char *keylist)
{
  BODY *pbody = NULL, *tmp_pbody = NULL;
  BODY *tmp_smime_pbody = NULL;
  BODY *tmp_pgp_pbody = NULL;
  int flags = msg->security;
  int i;

  if ((msg->security & SIGN) && !crypt_valid_passphrase (msg->security))
    return (-1);

  if ((msg->security & PGPINLINE) == PGPINLINE) {
    /* they really want to send it inline... go for it */
    if (!isendwin ())
      mutt_endwin _("Invoking PGP...");

    pbody = crypt_pgp_traditional_encryptsign (msg->content, flags, keylist);
    if (pbody) {
      msg->content = pbody;
      return 0;
    }

    /* otherwise inline won't work...ask for revert */
    if ((i =
         query_quadoption (OPT_PGPMIMEAUTO,
                           _
                           ("Message can't be sent inline.  Revert to using PGP/MIME?")))
        != M_YES) {
      mutt_error _("Mail not sent.");

      return -1;
    }

    /* go ahead with PGP/MIME */
  }

  if (!isendwin ())
    mutt_endwin (NULL);

  tmp_smime_pbody = msg->content;
  tmp_pgp_pbody = msg->content;

  if (msg->security & SIGN) {
    if (msg->security & APPLICATION_SMIME) {
      if (!(tmp_pbody = crypt_smime_sign_message (msg->content)))
        return -1;
      pbody = tmp_smime_pbody = tmp_pbody;
    }

    if ((msg->security & APPLICATION_PGP)
        && (!(flags & ENCRYPT) || option (OPTPGPRETAINABLESIG))) {
      if (!(tmp_pbody = crypt_pgp_sign_message (msg->content)))
        return -1;

      flags &= ~SIGN;
      pbody = tmp_pgp_pbody = tmp_pbody;
    }

    if ((msg->security & APPLICATION_SMIME)
        && (msg->security & APPLICATION_PGP)) {
      /* here comes the draft ;-) */
    }
  }


  if (msg->security & ENCRYPT) {
    if ((msg->security & APPLICATION_SMIME)) {
      if (!(tmp_pbody = crypt_smime_build_smime_entity (tmp_smime_pbody,
                                                        keylist))) {
        /* signed ? free it! */
        return (-1);
      }
      /* free tmp_body if messages was signed AND encrypted ... */
      if (tmp_smime_pbody != msg->content && tmp_smime_pbody != tmp_pbody) {
        /* detatch and dont't delete msg->content,
           which tmp_smime_pbody->parts after signing. */
        tmp_smime_pbody->parts = tmp_smime_pbody->parts->next;
        msg->content->next = NULL;
        body_list_wipe(&tmp_smime_pbody);
      }
      pbody = tmp_pbody;
    }

    if ((msg->security & APPLICATION_PGP)) {
      if (!(pbody = crypt_pgp_encrypt_message (tmp_pgp_pbody, keylist,
                                               flags & SIGN))) {

        /* did we perform a retainable signature? */
        if (flags != msg->security) {
          /* remove the outer multipart layer */
          tmp_pgp_pbody = mutt_remove_multipart (tmp_pgp_pbody);
          /* get rid of the signature */
          body_list_wipe(&tmp_pgp_pbody->next);
        }

        return (-1);
      }

      /* destroy temporary signature envelope when doing retainable 
       * signatures.

       */
      if (flags != msg->security) {
        tmp_pgp_pbody = mutt_remove_multipart (tmp_pgp_pbody);
        body_list_wipe(&tmp_pgp_pbody->next);
      }
    }
  }

  if (pbody)
    msg->content = pbody;

  return 0;
}


int crypt_query (BODY * m)
{
  int t = 0;

  if (!m)
    return 0;

  if (m->type == TYPEAPPLICATION) {
    t |= mutt_is_application_pgp (m);

    t |= mutt_is_application_smime (m);
    if (t && m->goodsig)
      t |= GOODSIGN;
    if (t && m->badsig)
      t |= BADSIGN;
  }
  else if (m->type == TYPETEXT) {
    t |= mutt_is_application_pgp (m);
    if (t && m->goodsig)
      t |= GOODSIGN;
  }

  if (m->type == TYPEMULTIPART) {
    t |= mutt_is_multipart_encrypted (m);
    t |= mutt_is_multipart_signed (m);

    if (t && m->goodsig)
      t |= GOODSIGN;
  }

  if (m->type == TYPEMULTIPART || m->type == TYPEMESSAGE) {
    BODY *p;
    int u, v, w;

    u = m->parts ? 0xffffffff : 0;      /* Bits set in all parts */
    w = 0;                      /* Bits set in any part  */

    for (p = m->parts; p; p = p->next) {
      v = crypt_query (p);
      u &= v;
      w |= v;
    }
    t |= u | (w & ~GOODSIGN);

    if ((w & GOODSIGN) && !(u & GOODSIGN))
      t |= PARTSIGN;
  }

  return t;
}


int crypt_write_signed (BODY * a, STATE * s, const char *tempfile)
{
  FILE *fp;
  int c;
  short hadcr;
  size_t bytes;

  if (!(fp = safe_fopen (tempfile, "w"))) {
    mutt_perror (tempfile);
    return -1;
  }

  fseeko (s->fpin, a->hdr_offset, 0);
  bytes = a->length + a->offset - a->hdr_offset;
  hadcr = 0;
  while (bytes > 0) {
    if ((c = fgetc (s->fpin)) == EOF)
      break;

    bytes--;

    if (c == '\r')
      hadcr = 1;
    else {
      if (c == '\n' && !hadcr)
        fputc ('\r', fp);

      hadcr = 0;
    }

    fputc (c, fp);

  }
  fclose (fp);

  return 0;
}



void convert_to_7bit (BODY * a)
{
  while (a) {
    if (a->type == TYPEMULTIPART) {
      if (a->encoding != ENC7BIT) {
        a->encoding = ENC7BIT;
        convert_to_7bit (a->parts);
      }
      else if (option (OPTPGPSTRICTENC))
        convert_to_7bit (a->parts);
    }
    else if (a->type == TYPEMESSAGE &&
             m_strcasecmp(a->subtype, "delivery-status")) {
      if (a->encoding != ENC7BIT)
        mutt_message_to_7bit (a, NULL);
    }
    else if (a->encoding == ENC8BIT)
      a->encoding = ENCQUOTEDPRINTABLE;
    else if (a->encoding == ENCBINARY)
      a->encoding = ENCBASE64;
    else if (a->content && a->encoding != ENCBASE64 &&
             (a->content->from || (a->content->space &&
                                   option (OPTPGPSTRICTENC))))
      a->encoding = ENCQUOTEDPRINTABLE;
    a = a->next;
  }
}


void crypt_extract_keys_from_messages (HEADER * h)
{
  int i;
  char tempfname[_POSIX_PATH_MAX], *mbox;
  address_t *tmp = NULL;
  FILE *fpout;

  mutt_mktemp (tempfname);
  if (!(fpout = safe_fopen (tempfname, "w"))) {
    mutt_perror (tempfname);
    return;
  }

  set_option (OPTDONTHANDLEPGPKEYS);

  if (!h) {
    for (i = 0; i < Context->vcount; i++) {
      if (Context->hdrs[Context->v2r[i]]->tagged) {
        mutt_parse_mime_message (Context, Context->hdrs[Context->v2r[i]]);
        if (Context->hdrs[Context->v2r[i]]->security & ENCRYPT &&
            !crypt_valid_passphrase (Context->hdrs[Context->v2r[i]]->
                                     security)) {
          fclose (fpout);
          break;
        }

        if (Context->hdrs[Context->v2r[i]]->security & APPLICATION_PGP) {
          mutt_copy_message (fpout, Context, Context->hdrs[Context->v2r[i]],
                             M_CM_DECODE | M_CM_CHARCONV, 0);
          fflush (fpout);

          mutt_endwin (_("Trying to extract PGP keys...\n"));
          crypt_pgp_invoke_import (tempfname);
        }

        if (Context->hdrs[Context->v2r[i]]->security & APPLICATION_SMIME) {
          if (Context->hdrs[Context->v2r[i]]->security & ENCRYPT)
            mutt_copy_message (fpout, Context, Context->hdrs[Context->v2r[i]],
                               M_CM_NOHEADER | M_CM_DECODE_CRYPT
                               | M_CM_DECODE_SMIME, 0);
          else
            mutt_copy_message (fpout, Context,
                               Context->hdrs[Context->v2r[i]], 0, 0);
          fflush (fpout);

          if (Context->hdrs[Context->v2r[i]]->env->from)
            tmp = mutt_expand_aliases (h->env->from);
          else if (Context->hdrs[Context->v2r[i]]->env->sender)
            tmp = mutt_expand_aliases (Context->hdrs[Context->v2r[i]]
                                       ->env->sender);
          mbox = tmp ? tmp->mailbox : NULL;
          if (mbox) {
            mutt_endwin (_("Trying to extract S/MIME certificates...\n"));
            crypt_smime_invoke_import (tempfname, mbox);
            tmp = NULL;
          }
        }

        rewind (fpout);
      }
    }
  }
  else {
    mutt_parse_mime_message (Context, h);
    if (!(h->security & ENCRYPT && !crypt_valid_passphrase (h->security))) {
      if (h->security & APPLICATION_PGP) {
        mutt_copy_message (fpout, Context, h, M_CM_DECODE | M_CM_CHARCONV, 0);
        fflush (fpout);
        mutt_endwin (_("Trying to extract PGP keys...\n"));
        crypt_pgp_invoke_import (tempfname);
      }

      if (h->security & APPLICATION_SMIME) {
        if (h->security & ENCRYPT)
          mutt_copy_message (fpout, Context, h, M_CM_NOHEADER
                             | M_CM_DECODE_CRYPT | M_CM_DECODE_SMIME, 0);
        else
          mutt_copy_message (fpout, Context, h, 0, 0);

        fflush (fpout);
        if (h->env->from)
          tmp = mutt_expand_aliases (h->env->from);
        else if (h->env->sender)
          tmp = mutt_expand_aliases (h->env->sender);
        mbox = tmp ? tmp->mailbox : NULL;
        if (mbox) {             /* else ? */
          mutt_message (_("Trying to extract S/MIME certificates...\n"));
          crypt_smime_invoke_import (tempfname, mbox);
        }
      }
    }
  }

  fclose (fpout);
  if (isendwin ())
    mutt_any_key_to_continue (NULL);

  mutt_unlink (tempfname);

  unset_option (OPTDONTHANDLEPGPKEYS);
}



int crypt_get_keys (HEADER * msg, char **keylist)
{
  /* Do a quick check to make sure that we can find all of the encryption
   * keys if the user has requested this service.
   */

  set_option (OPTPGPCHECKTRUST);

  *keylist = NULL;

  if (msg->security & ENCRYPT) {
    if (msg->security & APPLICATION_PGP) {
      if ((*keylist = crypt_pgp_findkeys (msg->env->to, msg->env->cc,
                                          msg->env->bcc)) == NULL)
        return (-1);
      unset_option (OPTPGPCHECKTRUST);
    }
    if (msg->security & APPLICATION_SMIME) {
      if ((*keylist = crypt_smime_findkeys (msg->env->to, msg->env->cc,
                                            msg->env->bcc)) == NULL)
        return (-1);
    }
  }

  return (0);
}



static void crypt_fetch_signatures (BODY ***signatures, BODY * a, int *n)
{
  for (; a; a = a->next) {
    if (a->type == TYPEMULTIPART)
      crypt_fetch_signatures (signatures, a->parts, n);
    else {
      if ((*n % 5) == 0)
        p_realloc(signatures, *n + 6);

      (*signatures)[(*n)++] = a;
    }
  }
}


/*
 * This routine verifies a  "multipart/signed"  body.
 */

int mutt_signed_handler (BODY * a, STATE * s)
{
  char tempfile[_POSIX_PATH_MAX];
  char *protocol;
  int protocol_major = TYPEOTHER;
  char *protocol_minor = NULL;

  BODY *b = a;
  BODY **signatures = NULL;
  int sigcnt = 0;
  int i;
  short goodsig = 1;
  int rc = 0;

  protocol = parameter_getval(a->parameter, "protocol");
  a = a->parts;

  /* extract the protocol information */

  if (protocol) {
    char major[STRING];
    char *t;

    if ((protocol_minor = strchr (protocol, '/')))
      protocol_minor++;

    m_strcpy(major, sizeof(major), protocol);
    if ((t = strchr (major, '/')))
      *t = '\0';

    protocol_major = mutt_check_mime_type (major);
  }

  /* consistency check */

  if (!(a && a->next && a->next->type == protocol_major &&
        !m_strcasecmp(a->next->subtype, protocol_minor))) {
    state_attach_puts (_("[-- Error: "
                         "Inconsistent multipart/signed structure! --]\n\n"),
                       s);
    return mutt_body_handler (a, s);
  }


  if (protocol_major == TYPEAPPLICATION
      && !m_strcasecmp(protocol_minor, "pgp-signature"));
  else if (protocol_major == TYPEAPPLICATION
           && !(m_strcasecmp(protocol_minor, "x-pkcs7-signature")
                && m_strcasecmp(protocol_minor, "pkcs7-signature")));
  else if (protocol_major == TYPEMULTIPART
           && !m_strcasecmp(protocol_minor, "mixed"));
  else {
    state_printf (s, _("[-- Error: "
                       "Unknown multipart/signed protocol %s! --]\n\n"),
                  protocol);
    return mutt_body_handler (a, s);
  }

  if (s->flags & M_DISPLAY) {

    crypt_fetch_signatures (&signatures, a->next, &sigcnt);

    if (sigcnt) {
      mutt_mktemp (tempfile);
      if (crypt_write_signed (a, s, tempfile) == 0) {
        for (i = 0; i < sigcnt; i++) {
          if (signatures[i]->type == TYPEAPPLICATION
              && !m_strcasecmp(signatures[i]->subtype, "pgp-signature")) {
            if (crypt_pgp_verify_one (signatures[i], s, tempfile) != 0)
              goodsig = 0;

            continue;
          }

          if (signatures[i]->type == TYPEAPPLICATION
          && (!m_strcasecmp(signatures[i]->subtype, "x-pkcs7-signature")
               || !m_strcasecmp(signatures[i]->subtype, "pkcs7-signature")))
          {
            if (crypt_smime_verify_one (signatures[i], s, tempfile) != 0)
              goodsig = 0;

            continue;
          }

          state_printf (s, _("[-- Warning: "
                             "We can't verify %s/%s signatures. --]\n\n"),
                        TYPE (signatures[i]), signatures[i]->subtype);
        }
      }

      mutt_unlink (tempfile);

      b->goodsig = goodsig;
      b->badsig = !goodsig;

      /* Now display the signed body */
      state_attach_puts (_("[-- The following data is signed --]\n\n"), s);


      p_delete(&signatures);
    }
    else
      state_attach_puts (_("[-- Warning: Can't find any signatures. --]\n\n"),
                         s);
  }

  rc = mutt_body_handler (a, s);

  if (s->flags & M_DISPLAY && sigcnt)
    state_attach_puts (_("\n[-- End of signed data --]\n"), s);

  return (rc);
}
